/*
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2011, 2015.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 */

(function(){
dojo.provide("com.ibm.rdm.web.server.cache.ProcessOperationsCache");

dojo.require("com.ibm.rdm.web.common.ProjectCache");
dojo.require("com.ibm.rdm.web.model.resources.ResourceManager");
dojo.require("com.ibm.rdm.web.model.types.TypesManager");
dojo.require("com.ibm.rdm.web.server.cache.ModifyArtifactPermissions");
dojo.require("com.ibm.rdm.web.server.client.RDMClient");
dojo.require("com.ibm.rdm.web.notifications.ChangeNotifier");
dojo.require("com.ibm.rdm.web.commenting.EventTopics");

var ProjectCache = com.ibm.rdm.web.common.ProjectCache;
var RDMClient = com.ibm.rdm.web.server.client.RDMClient;
var ResourceManager = com.ibm.rdm.web.model.resources.ResourceManager;
var ModifyArtifactPermissions = com.ibm.rdm.web.server.cache.ModifyArtifactPermissions;
var TypesManager = com.ibm.rdm.web.model.types.TypesManager;
var ChangeNotifier = com.ibm.rdm.web.notifications.ChangeNotifier;

var MODIFY_LINK_TYPE = "modifyLinkType";
var CLEAR_SUSPECT_STATE = "clearSuspectState";
var SET_SUSPECT_STATE = "setSuspectState";


var mapProjectUriToProjectPermissionsTracker = {};

com.ibm.rdm.web.server.cache.ProcessOperationsCache = {

	// summary:
	//		This singleton is responsible for providing "discovery" calls to the
	//		"Team Permissions Discovery" service.  Although this class doesn't
	//		explicitly cache anything, it does make use of the RMDClient which
	//		means that, in the end, cached values area being used.

	///////////////////////////////////////////////////////////////////////////
	// Public fields

	PERMISSION_TYPES: {
		Create: "create",
		Modify: "modify",
		Delete: "delete"
	},

	///////////////////////////////////////////////////////////////////////////
	// Public API

	getFolderProcessAreaURI: function(folderURI, /*Boolean?*/skipCache) {
		// TODO (Note by SPC) ...
		// We should change the following getFolder call to avoid needing to
		// pass skipCache as true.  For some reason the following fails unless
		// we pass as true ...
		// Open an folder
		// (1) Use grid context menu to ask for Assign Team Ownership for
		//     one of the folder's artifacts and Cancel.
		// (2) Use the folder's context menu to ask for Assign Team Ownership
		//     for the folder, set it to a new Team Area and click on OK.
		// (3) Use grid context menu to ask for Assign Team Ownership for
		//     the artifact from (1)
		// The end result is that the console.debug found in this function
		// would be called.
		return ProjectCache.getFolder(
			folderURI,
			skipCache
		).addCallback(
			this,
			function(folderData) {
				// TODO remove this checking after we're able to use module asset folder correctly ...
				if (folderData.folder) {
					var processAreaURI = folderData.folder.processArea.xmlAttributes.resource;
					return processAreaURI;
				}
				else {
//					console.debug("ProcessOperationsCache.getFolderProcessAreaURI was called with folderURI set to ", folderURI);
//					console.trace();
//					console.dir(folderData);
				}
			}
		);
	},

	isSaveTeamSecuritySpecificationAllowed: function(processAreaURI, projectURI) {
		// summary:
		//		Return a deferred which, when called back, will pass a single
		//		boolean representing whether the user should be allowed to
		//		change the process area setting.
		return this._isSaveTeamSecuritySpecificationAllowed(processAreaURI, projectURI);
	},
	
	isQuickPrintAllowed: function(processAreaURI, projectURI) {
		// summary:
		//		Return a deferred which, when called back, will pass a single
		//		boolean representing whether the user should be allowed to
		//		use the Quick Print facility.
		return this._isQuickPrintAllowed(processAreaURI, projectURI);
	},	

	isDeleteInstanceOfArtifactTypeAllowed: function(
		resourceURI, projectURI, artifactTypeURI, processAreaURI,
		/*Boolean?*/doCheckPermissionsForFolderInsteadOfArtifact, /*Boolean*/skipCache)
	{
		// summary:
		//		Returns a deferred which, when fired, will call you back with a
		//		single boolean argument representing whether you should be able
		//		to delete an artifact of type artifactTypeURI within folder
		//		folderURI.
		// description:
		//		This function is called for two different use cases:
		//		(1) Determine whether a "delete artifact" operation should be
		//		    permitted.
		//		(2) Determine whether a "move artifact" operation should be
		//		    permitted.
		//		In the case of (1), we consult the passed processAreaURI when
		//		performing the Delete permission check.
		//		In the case of (2), we consult the processAreaURI that is
		//		currently assigned to the resource's containing folder when
		//		performing the Delete permission check.
		// processAreaURI:
		//		The process area associate with the resourceURI artifact.
		// doCheckPermissionsForFolderInsteadOfArtifact:
		//		Only pass as true if you're attempting to perform a "move
		//		artifact" permission check.  Please note that for "move
		//		artifact", the user only be able to initiate the move if the
		//		user is able to delete an artifact of the artifact's type
		//		from the artifact's containing folder.
		var deferred = new dojo.Deferred();
		var performDeletePermissionCheckAndInvokeCallback =
			dojo.hitch(this, function(processAreaURI, objectType) {
				var artifactTypeActionId = objectType.getActionId();
				var isDeletedAllowedDeferred =
					this._isDeleteInstanceOfArtifactTypeAllowed(
						artifactTypeActionId, projectURI, processAreaURI
					);
				isDeletedAllowedDeferred.addCallback(
					this,
					function(isDeleteInstanceOfArtifactTypeAllowed){
						deferred.callback(isDeleteInstanceOfArtifactTypeAllowed);
					}
				);
			});
		var typeDeferred = TypesManager.getObjectType({
			typeURI: artifactTypeURI,
			projectURI: projectURI
		});
		if (doCheckPermissionsForFolderInsteadOfArtifact) {
			// The process area URI we need to check against is the one
			// assigned to the folder that contains the artifact.
			var objectType = null;
			var folderProcessAreaURI = null;
			var folderProcessAreaUriDeferred = new dojo.Deferred();
			var descriptionDeferred = ResourceManager.getResourceDescribe({
				resourceURI: resourceURI,
				skipCache: skipCache  // We can't be sure the folderURI is included unless we get the results fresh
			});
			typeDeferred.addCallback(this, function(type){
				objectType = type;
			});
			descriptionDeferred.addCallback(this, function(descriptionResponse){
				var folderURI = descriptionResponse.parentResourceURI;
				if (folderURI) {
					this.getFolderProcessAreaURI(folderURI).addCallback(
						this,
						function(processAreaURI){
							folderProcessAreaURI = processAreaURI;
							folderProcessAreaUriDeferred.callback();
							return processAreaURI;
						}
					);
				}
			});
			var deferredsList = new dojo.DeferredList(
				[typeDeferred, folderProcessAreaUriDeferred]
			);
			deferredsList.addBoth(this, function(response){
				performDeletePermissionCheckAndInvokeCallback(
					folderProcessAreaURI, objectType
				);
				return response;
			});
		} else {
			// The process area URI we need to check against is the one
			// assigned to the artifact.
			typeDeferred.addCallback(this, function(objectType){
				performDeletePermissionCheckAndInvokeCallback(
					processAreaURI,
					objectType);
			});
		}
		return deferred;
	},

	isCreateInstanceOfArtifactTypeAllowed: function(projectURI, folderURI, artifactTypeURI,
													/*String?*/artifactTypeActionId,
													/*String?*/processAreaURI) {
		// summary:
		//		Returns a deferred which, when fired, will call you back with a
		//		single boolean argument representing whether you should be able
		//		to create an artifact of type artifactTypeURI within folder
		//		folderURI.
		// artifactTypeActionId:
		//		The action ID associated with the passed artifactTypeURI.
		//		This field is optional, but the caller should provide it if it's
		//		readily available (as this function will have to invoke additional
		//		deferreds if it's not passed).
		// processAreaURI:
		//		The process area URI that's associated with the passed folderURI.
		//		This field is optional, but the caller should provide it if it's
		//		readily available (as this function will have to invoke additional
		//		deferreds if it's not passed).
		var deferred = new dojo.Deferred();
		var performPermissionsCheck = dojo.hitch(
			this,
			function(artifactTypeActionId, processAreaURI) {
				var isCreateInstanceAllowedDeferred =
					this._isCreateInstanceOfArtifactTypeAllowed(
						artifactTypeActionId,
						projectURI,
						(processAreaURI || projectURI)
					);
				isCreateInstanceAllowedDeferred.then(
					function(isCreateInstanceOfArtifactTypeAllowed){
						deferred.callback(isCreateInstanceOfArtifactTypeAllowed);
					}
				);
			});
		var getProcessAreaURIAndPerformPermissionsCheck = dojo.hitch(
			this,
			function(artifactTypeActionId) {
				if (processAreaURI) {
					performPermissionsCheck(artifactTypeActionId, processAreaURI);
				} else {
					this.getFolderProcessAreaURI(folderURI).then(
						function(processAreaURI){
							performPermissionsCheck(artifactTypeActionId, processAreaURI);
						}
					);
				}
			}
		);
		if (artifactTypeActionId) {
			getProcessAreaURIAndPerformPermissionsCheck(artifactTypeActionId);
		} else {
			TypesManager.getObjectType({
				typeURI: artifactTypeURI,
				projectURI: projectURI
			}).then(
				function(objectType){
					var artifactTypeActionId = objectType.getActionId();
					getProcessAreaURIAndPerformPermissionsCheck(artifactTypeActionId);
				}
			);
		}
		return deferred;
	},

	isCreateFolderAllowed: function(folderProcessAreaURI, projectURI) {
		// summary:
		//		Returns a deferred which, when fired, will call you back with a
		//		single boolean argument representing whether you should be able
		//		to create new subfolders (under a folder whose process area
		//		assignment is folderProcessAreaURI).
		var deferred = new dojo.Deferred();
		var getPermissionsDeferred = this._getFolderPermissions(
			folderProcessAreaURI, projectURI
		);
		getPermissionsDeferred.addCallback(this, function(json){
			deferred.callback(json.isCreateAllowed);
		});
		return deferred;
	},

	isDeleteFolderAllowed: function(folderProcessAreaURI, projectURI) {
		// summary:
		//		Returns a deferred which, when fired, will call you back with a
		//		single boolean argument representing whether you should be able
		//		to delete a folder whose process area assignment is
		//		folderProcessAreaURI.
		var deferred = new dojo.Deferred();
		var getPermissionsDeferred = this._getFolderPermissions(
			folderProcessAreaURI, projectURI
		);
		getPermissionsDeferred.addCallback(this, function(json){
			deferred.callback(json.isDeleteAllowed);
		});
		return deferred;
	},

	isModifyFolderAllowed: function(folderProcessAreaURI, projectURI) {
		// summary:
		//		Returns a deferred which, when fired, will call you back with a
		//		single boolean argument representing whether you should be able
		//		to modify a folder whose process area assignment is
		//		folderProcessAreaURI.
		var deferred = new dojo.Deferred();
		var getPermissionsDeferred = this._getFolderPermissions(
			folderProcessAreaURI, projectURI
		);
		getPermissionsDeferred.addCallback(this, function(json){
			deferred.callback(json.isModifyAllowed);
		});
		return deferred;
	},
	
	isReleaseLocksAllowed: function(processAreaURI, projectURI) {
		// summary:
		//		Returns a deferred which, when fired, will call you back with a
		//		single boolean argument representing whether you should be able
		//		to release a lock whose process area assignment is
		//		processAreaURI.
		var deferred = new dojo.Deferred();
		var getPermissionsDeferred = this._getModifyLocksPermissions(
			processAreaURI, projectURI
		);
		getPermissionsDeferred.addCallback(this, function(json){
			deferred.callback(json.isReleaseAllowed);
		});
		return deferred;
	},

	getModifyArtifactPermissions: function(artifactTypeActionId, processAreaURI, projectURI) {
		// summary:
		//		Returns a deferred which, when fired, will call you back with a
		//		single argument which will be an instance of
		//		ModifyArtifactPermissions.  This callback argument enables you
		//		to inquire about whether the artifact in question is editable
		//		and, for all of its attributes, which of these attributes are
		//		editable.
		
		return this._getModifyArtifactPermissions(
			artifactTypeActionId,
			projectURI,
			(processAreaURI || projectURI)
		);
	},

	getMultiModifyArtifactPermissions: function(idsAndProcessAreas, projectURI) {
		// summary:
		//		Returns information about the modify permissions for the passed
		//		'idsAndProcessAreas'.
		// idsAndProcessAreas:
		//		An array of JavaScript objects formatted as following:
		//			[{ "actionId": actionId, "processArea": processAreaURI }]
		// returns:
		//		A DeferredList which, when fired, will contain Deferred instances, each
		//		of which will pass a single parameter representing a ModifyArtifactPermissions
		//		instance.  The order of the deferreds appearing in this list will match the
		//		order of the passed 'idsAndProcessAreas'.  This return value allows you to
		//		inquire about whether the artifacts in question are editable and which of
		//		their attributes are editable.
		
		return this._getMultiArtifactPermissions(
			idsAndProcessAreas, projectURI, this.PERMISSION_TYPES.Modify
		);
	},
	
	getCreateArtifactTypeInstancesInFolderAllowed: function(objectTypes, folderURI, projectURI) {
		// summary:
		//		Returns information about the create permissions for the passed
		//		'objectTypes' with regard to the passed 'folderURI'.
		// returns:
		//		A DeferredList which, when fired, will contain Deferred instances, each
		//		of which will pass a single parameter representing a boolean (set to true only
		//		when you're able to create an instance of that type in the passed folder).
		//		The order of the deferreds appearing in this list will match the order of the
		//		passed 'objectTypes'.  This return value allows you to inquire about whether
		//		the artifacts in question are "creatable" within the passed folder.

		return this.getFolderProcessAreaURI(folderURI).addCallback(this, function(processAreaURI){
			var pairs = [],
				i;
			for (i = 0; i < objectTypes.length; i++) {
				pairs.push({
					actionId: objectTypes[i].getActionId(),
					processArea: processAreaURI
				});
			}
			return this._getMultiArtifactPermissions(
				pairs, projectURI, this.PERMISSION_TYPES.Create
			);
		});
	},
	
	//#68914 Attributes added to artifact are shown but can't be edited
	artifactTypeChangedNotify: function(eventParams, tracker) {
		// summary: called from ProjectPermissionsTracker when artifact attributes are changed
		var projectURI = eventParams.notifierParams.projectURI;
		var permissionsTracker = this._getOrCreateProjectPermissionsTracker(projectURI);
		if (permissionsTracker === tracker) { 
			var artifactTypeActionId = eventParams.notifierParams.artifactTypeActionId;
			if (artifactTypeActionId) {
				//look for artifact type, delete it from the cache
				for(var propKey in tracker.mapModifyArtifactTypeToDeferred) {
					var i = propKey.indexOf(artifactTypeActionId);
					if (i > -1) {
						tracker.mapModifyArtifactTypeToDeferred[propKey] = null;
						break;
					}
				}
			}	
		}	
	},

	///////////////////////////////////////////////////////////////////////////
	// Link create, delete, modify permissions

	getPermissibleLinkTypesForCreation: function(projectURI, processAreaURI) {
		// summary:
		//		Returns a deferred which, when fired, will call you back with a
		//		map of links by actionId and whether not creation of the link
		//		is allowed.
		return this._getPermissibleLinkTypesForCreation(
			projectURI,
			(processAreaURI || projectURI)
		);
	},

	isCreateLinkAllowed: function(projectURI, linkTypeActionId, processAreaURI) {
		// summary:
		//		Returns a deferred which, when fired, returns whether or not a
		//		link of the given type can be created for the given processArea.
		var deferred = new dojo.Deferred();
		if (!linkTypeActionId) {
			deferred.callback(true);
		} else {
			this._isCreateLinkAllowed(
				projectURI,
				linkTypeActionId,
				(processAreaURI || projectURI)
			).then(
				function(isCreateAllowed){
					deferred.callback(isCreateAllowed);
				}
			);
		}
		return deferred;
	},

	isDeleteLinkAllowed: function(projectURI, linkTypeActionId, processAreaURI) {
		// summary:
		//		Returns a deferred which, when fired, answers whether or not the
		//		given linkType can be deleted for the given project and
		//		processAreaURI.
		var deferred = new dojo.Deferred();
		if (!linkTypeActionId) {
			deferred.callback(true);
		} else {
			this._isDeleteLinkAllowed(
				projectURI,
				linkTypeActionId,
				(processAreaURI || projectURI)
			).then(
				function(isDeleteAllowed){
					deferred.callback(isDeleteAllowed);
				}
			);
		}
		return deferred;
	},

	isModifyLinkTypeAllowed: function(projectURI, linkTypeActionId, processAreaURI) {
		// summary:
		return this._isModifyLinkAllowed(
			projectURI,
			linkTypeActionId,
			(processAreaURI || projectURI),
			MODIFY_LINK_TYPE
		);
	},
	
	isClearSuspectLinkAllowed: function(projectURI, linkTypeActionId, processAreaURI) {
		// summary:
		return this._isModifyLinkAllowed(
			projectURI,
			linkTypeActionId,
			(processAreaURI || projectURI),
			CLEAR_SUSPECT_STATE
		);
	},
	
	isSetSuspectLinkAllowed: function(projectURI, linkTypeActionId, processAreaURI) {
		// summary:
		return this._isModifyLinkAllowed(
			projectURI,
			linkTypeActionId,
			(processAreaURI || projectURI),
			SET_SUSPECT_STATE
		);
	},

	///////////////////////////////////////////////////////////////////////////
	// Private API

	_createActionPath: function(actionId, permissionType) {
		switch(permissionType) {
		case this.PERMISSION_TYPES.Create:
			return "create/specificType/" + actionId;
		case this.PERMISSION_TYPES.Modify:
			return "modify/specificType/" + actionId;
		}		
	},

	_getMultiArtifactPermissions: function(idsAndProcessAreas, projectURI, permissionType) {
		// summary:
		//		Returns information about the create or modify permissions for the passed
		//		'idsAndProcessAreas'.
		// idsAndProcessAreas:
		//		An array of JavaScript objects formatted as following:
		//			[{ "actionId": actionId, "processArea": processAreaURI }]
		// permissionType:
		//		Must correspond to one of the values found in this.PERMISSION_TYPES
		//		(currently this function only supports Create and Modify [not Delete]).
		// returns:
		//		A DeferredList which, when fired, will contain Deferred instances, each
		//		of which will pass a single parameter representing either a
		//		ModifyArtifactPermissions instance (in the case of permissionType being
		//		Modify) or a boolean (in the case of permissionType being Create).
		
		var permissionsTracker = this._getOrCreateProjectPermissionsTracker(projectURI);
		var operationsDeferred;
		var operationsDeferredList;
		
		// Temporary arrays for building the POST
		var permissionsArray   = [];
		var missingIdsAndAreas = [];
		var missingCallbacks   = [];
		var knownCallbacks     = [];
		var returnCallbacks    = [];
		var permissionsObject  = {};
		var formattedRequest   = {};
		var tempXhrArgs        = {};
		var parseFunction;
		var tempDeferred;
		
		// Loop incrementers:
		//	i is for the main loop
		//  j maintains missingCallbacks' index
		//  k maintains knownCallbacks' index
		var i = 0;
		var j = 0;
		var k = 0;

		// We want each actionId to have its own deferred, so we only request permissions on new
		// actionIds without server calls already in flight.
		for (i = 0; i < idsAndProcessAreas.length; i+=1) {
		
			switch(permissionType) {
			case this.PERMISSION_TYPES.Create:
				tempDeferred = permissionsTracker.getCreateInstanceOfArtifactTypeAllowedDeferred(
					idsAndProcessAreas[i].processArea, idsAndProcessAreas[i].actionId
				);
				break;
			case this.PERMISSION_TYPES.Modify:
				tempDeferred = permissionsTracker.getModifyArtifactPermissionsDeferred(
					idsAndProcessAreas[i].processArea, idsAndProcessAreas[i].actionId
				);
				break;
			}
			
			// Keep the order of our original IDs intact, because otherwise the calling function has
			// no way to tell which permissions match with which IDs.
			if (!tempDeferred) {
				tempDeferred = new dojo.Deferred();
				
				missingIdsAndAreas.push(idsAndProcessAreas[i]);
				missingCallbacks.push(tempDeferred);

				returnCallbacks.push(missingCallbacks[j]);
				++j;
			}
			else {
				knownCallbacks.push(tempDeferred);
				
				returnCallbacks.push(knownCallbacks[k]);
				++k;
			}
		}
		
		// Begin building our request, closely following the format of this._getModifyArtifactPermissions
		dojo.forEach(missingIdsAndAreas, dojo.hitch(this, function(singleIdAndArea){

			// Build the JSON object for our request
			permissionsArray.push({
					action: [{"actionPath": this._createActionPath(singleIdAndArea.actionId, permissionType)}],
					processArea: singleIdAndArea.processArea
			});
		}));
		
		permissionsObject = {operation: "com.ibm.rrs.team.saveArtifact", 
							permissions: permissionsArray};
		
		// Format our request data
		tempXhrArgs = {
			headers:{"Accept": "text/json", "net.jazz.jfs.owning-context": projectURI, "Content-Type": "text/json"},
			handleAs: "json"
		};
		
		formattedRequest = {
			uri: com.ibm.rdm.web.FRONTING_SERVICE_OPERATIONS,
			skipCache : false,
			xhrArgs: tempXhrArgs,
			postData: dojo.toJson(permissionsObject)
		};	

		// Find all permissions pertaining to each actionId and separate them so each actionId has its own
		// ModifyArtifactPermissions object, and return the array.
		parseFunction = dojo.hitch(this, function(response) {
			// Construct our individual response object for each actionId, updating
			// only our missing ones.  This will in turn update the callbacks
			// in our "master list" returnCallbacks which includes already cached permissions.
			switch(permissionType) {
			case this.PERMISSION_TYPES.Create:
				for (var i = 0; i < response.doc.permissions.length; i++) {
					missingCallbacks[i].callback(
						response.doc.permissions[i].action[0].allowed === "true"
					);
				}
				break;
			case this.PERMISSION_TYPES.Modify:
				for (var i = 0; i < response.doc.permissions.length; i++) {
					missingCallbacks[i].callback(
						new ModifyArtifactPermissions({
							"license-info": response.doc["license-info"],
							"operation": response.doc["operation"],
							"permissions": [response.doc.permissions[i]]
						})
					);
				}
				break;
			}
		});
		
		formattedRequest["parser"] = parseFunction;

		// Fire our request.  Since we have individual callbacks for each permissions object,
		// we do not need to return operationsDeferred or add it to our deferredList.
		// The operationsDeferredList will ensure nothing will return before the rest are done.
		//
		// Only call if there are cache misses
		if (missingCallbacks.length) {
			operationsDeferred = RDMClient.doPost(formattedRequest);
		}
		
		operationsDeferredList = new dojo.DeferredList(returnCallbacks);

		// Set our missing deferreds, avoid subsequent unnecessary server calls for the same stuff by
		// keeping the known deferreds intact.
		switch(permissionType){
		case this.PERMISSION_TYPES.Create:
			for (i = 0; i < missingIdsAndAreas.length; i+=1) {
				permissionsTracker.setCreateInstanceOfArtifactTypeAllowedDeferred(
					missingIdsAndAreas[i].processArea,
					missingIdsAndAreas[i].actionId,
					missingCallbacks[i]
				);
			}
			break;
		case this.PERMISSION_TYPES.Modify:
			for (i = 0; i < missingIdsAndAreas.length; i+=1) {
				permissionsTracker.setModifyArtifactPermissionsDeferred(
					missingIdsAndAreas[i].processArea,
					missingIdsAndAreas[i].actionId,
					missingCallbacks[i]
				);
			}
			break;
		}
		
		return operationsDeferredList;
	},

	_getOrCreateProjectPermissionsTracker: function(projectURI) {
		var permissionsTracker = mapProjectUriToProjectPermissionsTracker[projectURI];
		if (!permissionsTracker) {
			permissionsTracker = new com.ibm.rdm.web.server.cache._ProjectPermissionsTracker();
			mapProjectUriToProjectPermissionsTracker[projectURI] = permissionsTracker;
		}
		return permissionsTracker;
	},


	isManualLockAcquireAllowed: function(processAreaURI, projectURI) {
		/*Operation ID: com.ibm.rrs.team.modifyLockOperation
		  Action ID: manualAcquire
		 */
		var permissionsTracker = this._getOrCreateProjectPermissionsTracker(projectURI);
		var operationsDeferred =
			permissionsTracker.getAcquireManualLockPermissionDeferred(
				processAreaURI
			);
		if (!operationsDeferred) {
			operationsDeferred = RDMClient.doPost({
				uri: com.ibm.rdm.web.FRONTING_SERVICE_OPERATIONS,
				skipCache: false,
				xhrArgs: {
					headers: {
						"Accept": "text/json",
						"net.jazz.jfs.owning-context": projectURI,
						"Content-Type": "text/json"
					},
					handleAs: "json"
				},
				postData:
					'{\n' +
					'   "operation": "com.ibm.rrs.team.modifyLockOperation",\n' +
					'   "permissions": [\n' +
					'       {\n' +
					'           "action": [\n' +
					'               {\n' +
					'                   "actionPath": "manualAcquire"\n' +
					'               }\n' +
					'           ],\n' +
					'           "processArea": "' + processAreaURI + '"\n' +
					'       }\n' +
					'   ]\n' +
					'}',
				parser: function(response) {
					return (response.doc.permissions[0].action[0].allowed === "true");
				}
			});
			permissionsTracker.setAcquireManualLockPermissionDeferred(
				processAreaURI, operationsDeferred
			);
		}
		
		return operationsDeferred;
		
	},

	_isSaveTeamSecuritySpecificationAllowed: function(processAreaURI, projectURI) {
		// summary:
		//		Return a deferred which, when called back, will pass a single
		//		boolean representing whether the user should be allowed to
		//		change the process area setting.
		var permissionsTracker = this._getOrCreateProjectPermissionsTracker(projectURI);
		var operationsDeferred =
			permissionsTracker.getSaveTeamSecuritySpecificationDeferred(
				processAreaURI
			);
		if (!operationsDeferred) {
			operationsDeferred = RDMClient.doPost({
				uri: com.ibm.rdm.web.FRONTING_SERVICE_OPERATIONS,
				skipCache: false,
				xhrArgs: {
					headers: {
						"Accept": "text/json",
						"net.jazz.jfs.owning-context": projectURI,
						"Content-Type": "text/json"
					},
					handleAs: "json"
				},
				postData:
					'{\n' +
					'	"operation": "com.ibm.rrs.team.saveTeamSecuritySpecification",\n' +
					'	"permissions": [\n' +
					'		{\n' +
					'			"action": [\n' +
					'				{\n' +
					'					"actionPath": "modify"\n' +
					'				}\n' +
					'			],\n' +
					'			"processArea": "' + processAreaURI + '"\n' +
					'		}\n' +
					'	]\n' +
					'}',
				parser: function(response) {
					return (response.doc.permissions[0].action[0].allowed === "true");
				}
			});
			permissionsTracker.setSaveTeamSecuritySpecificationDeferred(
				processAreaURI, operationsDeferred
			);
		}
		return operationsDeferred;
	},
	
	_isQuickPrintAllowed: function(processAreaURI, projectURI) {
		// summary:
		//		Return a deferred which, when called back, will pass a single
		//		boolean representing whether the user should be allowed to
		//		use the quick print facility.
		var permissionsTracker = this._getOrCreateProjectPermissionsTracker(projectURI);
		var operationsDeferred =
			permissionsTracker.getQuickPrintDeferred(
				processAreaURI
			);
		if (!operationsDeferred) {
			operationsDeferred = RDMClient.doPost({
				uri: com.ibm.rdm.web.FRONTING_SERVICE_OPERATIONS,
				skipCache: false,
				xhrArgs: {
					headers: {
						"Accept": "text/json",
						"net.jazz.jfs.owning-context": projectURI,
						"Content-Type": "text/json"
					},
					handleAs: "json"
				},
				postData:
					'{\n' +
					'	"operation": "com.ibm.rrs.team.modifyQuickPrintOperation",\n' +
					'	"permissions": [\n' +
					'		{\n' +
					'			"action": [\n' +
					'				{\n' +
					'					"actionPath": "release"\n' +
					'				}\n' +
					'			],\n' +
					'			"processArea": "' + processAreaURI + '"\n' +
					'		}\n' +
					'	]\n' +
					'}',
				parser: function(response) {
					return (response.doc.permissions[0].action[0].allowed === "true");
				}
			});
			permissionsTracker.setQuickPrintDeferred(
				processAreaURI, operationsDeferred
			);
		}
		return operationsDeferred;
	},	

	_getFolderPermissions: function(processAreaURI, projectURI) {
		var permissionsTracker = this._getOrCreateProjectPermissionsTracker(projectURI);
		var operationsDeferred = permissionsTracker.getFolderPermissionsDeferred(processAreaURI);
		if (!operationsDeferred) {
			operationsDeferred = RDMClient.doPost({
				uri: com.ibm.rdm.web.FRONTING_SERVICE_OPERATIONS,
				skipCache: false,
				xhrArgs: {
					headers: {
						"Accept": "text/json",
						"net.jazz.jfs.owning-context": projectURI,
						"Content-Type": "text/json"
					},
					handleAs: "json"
				},
				postData:
					'{\n' +
					'	"operation": "com.ibm.rrs.team.saveFolder",\n' +
					'	"permissions": [\n' +
					'		{\n' +
					'			"action": [\n' +
					'				{\n' +
					'					"actionPath": "create"\n' +
					'				},\n' +
					'				{\n' +
					'					"actionPath": "modify"\n' +
					'				},\n' +
					'				{\n' +
					'					"actionPath": "delete"\n' +
					'				}\n' +
					'			],\n' +
					'			"processArea": "' + processAreaURI + '"\n' +
					'		}\n' +
					'	]\n' +
					'}',
				parser: function(response) {
					var permissions = response.doc.permissions;
					return {
						isCreateAllowed: (permissions[0].action[0].allowed === "true"),
						isModifyAllowed: (permissions[0].action[1].allowed === "true"),
						isDeleteAllowed: (permissions[0].action[2].allowed === "true")
					};
				}
			});
			permissionsTracker.setFolderPermissionsDeferred(
				processAreaURI, operationsDeferred
			);
		}
		return operationsDeferred;
	},
	
	_getModifyLocksPermissions: function(processAreaURI, projectURI) {
		var permissionsTracker = this._getOrCreateProjectPermissionsTracker(projectURI);
		var operationsDeferred = permissionsTracker.getModifyLocksPermissionsDeferred(processAreaURI);
		if (!operationsDeferred) {
			operationsDeferred = RDMClient.doPost({
				uri: com.ibm.rdm.web.FRONTING_SERVICE_OPERATIONS,
				skipCache: false,
				xhrArgs: {
					headers: {
						"Accept": "text/json",
						"net.jazz.jfs.owning-context": projectURI,
						"Content-Type": "text/json"
					},
					handleAs: "json"
				},
				postData:
					'{\n' +
					'	"operation": "com.ibm.rrs.team.modifyLockOperation",\n' +
					'	"permissions": [\n' +
					'		{\n' +
					'			"action": [\n' +
					'				{\n' +
					'					"actionPath": "release"\n' +
					'				},\n' +
					'				{\n' +
					'					"actionPath": "acquirePersistent"\n' +
					'				}\n' +
					'			],\n' +
					'			"processArea": "' + processAreaURI + '"\n' +
					'		}\n' +
					'	]\n' +
					'}',
				parser: function(response) {
					var permissions = response.doc.permissions;
					return {
						isReleaseAllowed: (permissions[0].action[0].allowed === "true")
					};
				}
			});
			permissionsTracker.setModifyLocksPermissionsDeferred(
				processAreaURI, operationsDeferred
			);
		}
		return operationsDeferred;
	},
	
	_isDeleteInstanceOfArtifactTypeAllowed: function(
		artifactTypeActionId, projectURI, processAreaURI)
	{
		// summary:
		//		Return a deferred which, when called back, will pass a single
		//		boolean representing whether the user should be able to delete
		//		an instance of a particular artifact type (based on the passed
		//		artifactTypeActionId) within processAreaURI.
		var permissionsTracker = this._getOrCreateProjectPermissionsTracker(projectURI);
		var operationsDeferred =
			permissionsTracker.getDeleteInstanceOfArtifactTypeAllowedDeferred(
				processAreaURI, artifactTypeActionId
			);
		if (!operationsDeferred) {
			operationsDeferred = RDMClient.doPost({
				uri: com.ibm.rdm.web.FRONTING_SERVICE_OPERATIONS,
				skipCache: false,
				xhrArgs: {
					headers: {
						"Accept": "text/json",
						"net.jazz.jfs.owning-context": projectURI,
						"Content-Type": "text/json"
					},
					handleAs: "json"
				},
				postData:
					'{\n' +
					'	"operation": "com.ibm.rrs.team.saveArtifact",\n' +
					'	"permissions": [\n' +
					'		{\n' +
					'			"action": [\n' +
					'				{\n' +
					'					"actionPath": "delete/specificType/' + artifactTypeActionId + '"\n' +
					'				}\n' +
					'			],\n' +
					'			"processArea": "' + processAreaURI + '"\n' +
					'		}\n' +
					'	]\n' +
					'}',
				parser: function(response) {
					return (response.doc.permissions[0].action[0].allowed === "true");
				}
			});
			permissionsTracker.setDeleteInstanceOfArtifactTypeAllowedDeferred(
				processAreaURI, artifactTypeActionId, operationsDeferred
			);
		}
		return operationsDeferred;
	},

	_isCreateInstanceOfArtifactTypeAllowed: function(
		artifactTypeActionId, projectURI, processAreaURI)
	{
		// summary:
		//		Return a deferred which, when called back, will pass a single
		//		boolean representing whether the user should be able to delete
		//		an instance of a particular artifact type (based on the passed
		//		artifactTypeActionId) within processAreaURI.
		var permissionsTracker = this._getOrCreateProjectPermissionsTracker(projectURI);
		var operationsDeferred =
			permissionsTracker.getCreateInstanceOfArtifactTypeAllowedDeferred(
				processAreaURI, artifactTypeActionId
			);
		if (!operationsDeferred) {
			operationsDeferred = RDMClient.doPost({
				uri: com.ibm.rdm.web.FRONTING_SERVICE_OPERATIONS,
				skipCache: false,
				xhrArgs: {
					headers: {
						"Accept": "text/json",
						"net.jazz.jfs.owning-context": projectURI,
						"Content-Type": "text/json"
					},
					handleAs: "json"
				},
				postData:
					'{\n' +
					'	"operation": "com.ibm.rrs.team.saveArtifact",\n' +
					'	"permissions": [\n' +
					'		{\n' +
					'			"action": [\n' +
					'				{\n' +
					'					"actionPath": "create/specificType/' + artifactTypeActionId + '"\n' +
					'				}\n' +
					'			],\n' +
					'			"processArea": "' + processAreaURI + '"\n' +
					'		}\n' +
					'	]\n' +
					'}',
				parser: function(response) {
					return (response.doc.permissions[0].action[0].allowed === "true");
				}
			});
			permissionsTracker.setCreateInstanceOfArtifactTypeAllowedDeferred(
				processAreaURI, artifactTypeActionId, operationsDeferred
			);
		}
		return operationsDeferred;
	},

	_getModifyArtifactPermissions: function(
		artifactTypeActionId, projectURI, processAreaURI)
	{
		var permissionsTracker = this._getOrCreateProjectPermissionsTracker(projectURI);
		var operationsDeferred =
			permissionsTracker.getModifyArtifactPermissionsDeferred(
				processAreaURI, artifactTypeActionId
			);
		if (!operationsDeferred) {
			operationsDeferred = RDMClient.doPost({
				uri: com.ibm.rdm.web.FRONTING_SERVICE_OPERATIONS,
				skipCache: false,
				xhrArgs: {
					headers: {
						"Accept": "text/json",
						"net.jazz.jfs.owning-context": projectURI,
						"Content-Type": "text/json"
					},
					handleAs: "json"
				},
				postData:
					'{\n' +
					'	"operation": "com.ibm.rrs.team.saveArtifact",\n' +
					'	"permissions": [\n' +
					'		{\n' +
					'			"action": [\n' +
					'				{\n' +
					'					"actionPath": "modify/specificType/' + artifactTypeActionId + '"\n' +
					'				}\n' +
					'			],\n' +
					'			"processArea": "' + processAreaURI + '"\n' +
					'		}\n' +
					'	]\n' +
					'}',
				parser: function(response) {
					return new ModifyArtifactPermissions(response.doc);
				}
			});
			permissionsTracker.setModifyArtifactPermissionsDeferred(
				processAreaURI, artifactTypeActionId, operationsDeferred
			);
		}
		return operationsDeferred;
	},
	
	_isModifyLinkAllowed: function(
		projectURI, linkTypeActionId, processAreaURI, linkTypePermission) {
		// summary:
		//		Returns a deferred which, when fired, answers whether or not
		//		the given linkType can be modified for the given project and
		//		processAreaURI and linkTypePermission.
		var deferred = new dojo.Deferred();
		if (!linkTypeActionId) {
			deferred.callback(true);
		} else {
			var permissionsTracker = this._getOrCreateProjectPermissionsTracker(projectURI);
			var operationsDeferred =
				permissionsTracker.getModifyLinkDeferred(
					processAreaURI, linkTypeActionId, linkTypePermission
				);
			if (!operationsDeferred) {
				operationsDeferred = RDMClient.doPost({
					uri: com.ibm.rdm.web.FRONTING_SERVICE_OPERATIONS,
					skipCache: false,
					xhrArgs: {
						headers: {
							"Accept": "text/json",
							"net.jazz.jfs.owning-context": projectURI,
							"Content-Type": "text/json"
						},
						handleAs: "json"
					},
					postData:
						'{\n' +
						'	"operation": "com.ibm.rrs.team.saveLink",\n' +
						'	"permissions": [\n' +
						'		{\n' +
						'			"action": [\n' +
						'				{\n' +
						'					"actionPath": "modify/specificType/' + linkTypeActionId + '"\n' +
						'				}\n' +
						'			],\n' +
						'			"processArea": "' + processAreaURI + '"\n' +
						'		}\n' +
						'	]\n' +
						'}',
					parser: function(response) {
						var actions = response.doc.permissions[0].action;
						var isAllowed = true;
						for (var i=0; i<actions.length; i++) {
							var entry = actions[i];
							var actionPath = entry.actionPath;
							var lastSlashIndex = actionPath.lastIndexOf("/");
							var actionId = actionPath.substring(lastSlashIndex + 1);
							if (actionId === linkTypePermission) {
								isAllowed = entry.allowed === "true";
								break;
							}
						}
						return isAllowed;
					}
				});
				permissionsTracker.setModifyLinkDeferred(
					processAreaURI, linkTypeActionId, linkTypePermission, operationsDeferred
				);
			}
			operationsDeferred.then(
				function(isModifyAllowed) {
					deferred.callback(isModifyAllowed);
				}
			);
		}
		return deferred;
	},

	_getPermissibleLinkTypesForCreation: function(projectURI, processAreaURI) {
		// summary:
		//		Returns a deferred which, when fired, will call you back with a
		//		map of links by actionId and whether not creation of the link
		//		is allowed.
		var permissionsTracker = this._getOrCreateProjectPermissionsTracker(projectURI);
		var operationsDeferred =
			permissionsTracker.getPermissibleLinkTypesForCreationDeferred(
				processAreaURI
			);
		if (!operationsDeferred) {
			operationsDeferred = RDMClient.doPost({
				uri: com.ibm.rdm.web.FRONTING_SERVICE_OPERATIONS,
				skipCache: false,
				xhrArgs: {
					headers: {
						"Accept": "text/json",
						"net.jazz.jfs.owning-context": projectURI,
						"Content-Type": "text/json"
					},
					handleAs: "json"
				},
				postData:
					'{\n' +
					'	"operation": "com.ibm.rrs.team.saveLink",\n' +
					'	"permissions": [\n' +
					'		{\n' +
					'			"action": [\n' +
					'				{\n' +
					'					"actionPath": "create"\n' +
					'				}\n' +
					'			],\n' +
					'			"processArea": "' + processAreaURI + '"\n' +
					'		}\n' +
					'	]\n' +
					'}',
				parser: function(response) {
					var mapLinkTypesToAllowed = {};
					var actions = response.doc.permissions[0].action;
					for (var i = 0; i < actions.length; i++) {
						var entry = actions[i];
						var actionPath = entry.actionPath;
						var isAllowed = entry.allowed === "true";
						var lastSlashIndex = actionPath.lastIndexOf("/");
						var actionId = actionPath.substring(lastSlashIndex + 1);
						mapLinkTypesToAllowed[actionId] = isAllowed;
					}
					return mapLinkTypesToAllowed;
				}
			});
			permissionsTracker.setPermissibleLinkTypesForCreationDeferred(
				processAreaURI, operationsDeferred
			);
		}
		return operationsDeferred;
	},

	_isCreateLinkAllowed: function(projectURI, linkTypeActionId, processAreaURI) {
		var permissionsTracker = this._getOrCreateProjectPermissionsTracker(projectURI);
		var operationsDeferred =
			permissionsTracker.getCreateLinkDeferred(
				processAreaURI, linkTypeActionId
			);
		if (!operationsDeferred) {
			operationsDeferred = RDMClient.doPost({
				uri: com.ibm.rdm.web.FRONTING_SERVICE_OPERATIONS,
				skipCache: false,
				xhrArgs: {
					headers: {
						"Accept": "text/json",
						"net.jazz.jfs.owning-context": projectURI,
						"Content-Type": "text/json"
					},
					handleAs: "json"
				},
				postData:
					'{\n' +
					'	"operation": "com.ibm.rrs.team.saveLink",\n' +
					'	"permissions": [\n' +
					'		{\n' +
					'			"action": [\n' +
					'				{\n' +
					'					"actionPath": "create/specificType/' + linkTypeActionId + '"\n' +
					'				}\n' +
					'			],\n' +
					'			"processArea": "' + processAreaURI + '"\n' +
					'		}\n' +
					'	]\n' +
					'}',
				parser: function(response) {
					return (response.doc.permissions[0].action[0].allowed === "true");
				}
			});
			permissionsTracker.setCreateLinkDeferred(
				processAreaURI, linkTypeActionId, operationsDeferred
			);
		}
		return operationsDeferred;
	},

	_isDeleteLinkAllowed: function(projectURI, linkTypeActionId, processAreaURI) {
		var permissionsTracker = this._getOrCreateProjectPermissionsTracker(projectURI);
		var operationsDeferred =
			permissionsTracker.getDeleteLinkDeferred(
				processAreaURI, linkTypeActionId
			);
		if (!operationsDeferred) {
			operationsDeferred = RDMClient.doPost({
				uri: com.ibm.rdm.web.FRONTING_SERVICE_OPERATIONS,
				skipCache: false,
				xhrArgs: {
					headers: {
						"Accept": "text/json",
						"net.jazz.jfs.owning-context": projectURI,
						"Content-Type": "text/json"
					},
					handleAs: "json"
				},
				postData:
					'{\n' +
					'	"operation": "com.ibm.rrs.team.saveLink",\n' +
					'	"permissions": [\n' +
					'		{\n' +
					'			"action": [\n' +
					'				{\n' +
					'					"actionPath": "delete/specificType/' + linkTypeActionId + '"\n' +
					'				}\n' +
					'			],\n' +
					'			"processArea": "' + processAreaURI + '"\n' +
					'		}\n' +
					'	]\n' +
					'}',
				parser: function(response){
					return (response.doc.permissions[0].action[0].allowed === "true");
				}
			});
			permissionsTracker.setDeleteLinkDeferred(
				processAreaURI, linkTypeActionId, operationsDeferred
			);
		}
		return operationsDeferred;
	}
};


dojo.declare(
	"com.ibm.rdm.web.server.cache._ProjectPermissionsTracker",
	null,
{
	// summary:
	//		This is a utilitiy class that's used to allow for "session based
	//		permission checking" (as opposed to the default which would be
	//		"page based permission checking" based on our use of
	//		com.ibm.rdm.web.common.UpdateContext).

	mapSaveTeamSecurityToDeferred: null,
	mapFolderProcessAreaToDeferred: null,
	mapModifyLocksProcessAreaToDeferred: null,
	mapDeleteInstanceToDeferred: null,
	mapCreateInstanceToDeferred: null,
	mapModifyArtifactTypeToDeferred: null,
	mapModifyLinkInstanceToDeferred: null,
	mapPermissibleLinkTypesForCreationToDeferred: null,
	mapCreateLinkInstanceToDeferred: null,
	mapDeleteLinkInstanceToDeferred: null,
	mapAcquireManualLockDeferred: null,
	mapQuickPrintToDeferred: null,

	constructor: function() {
		this.mapSaveTeamSecurityToDeferred = {};
		this.mapFolderProcessAreaToDeferred = {};
		this.mapModifyLocksProcessAreaToDeferred = {};
		this.mapDeleteInstanceToDeferred = {};
		this.mapCreateInstanceToDeferred = {};
		this.mapModifyArtifactTypeToDeferred = {};
		this.mapModifyLinkInstanceToDeferred = {};
		this.mapPermissibleLinkTypesForCreationToDeferred = {};
		this.mapCreateLinkInstanceToDeferred = {};
		this.mapDeleteLinkInstanceToDeferred = {};
		this.mapAcquireManualLockDeferred = {};
		this.mapQuickPrintToDeferred = {};
		
		ChangeNotifier.addListener({
			eventTopics: [com.ibm.rdm.web.notifications.EventTopics.LINK_TYPE_ADDED],
			callbackFunction: this._onLinkTypeAddedNotify,
			callbackScope: this,
			neverDelayCallback: true
		});
		//#68914 Attributes added to artifact are shown but can't be edited
		ChangeNotifier.addListener({
			eventTopics: [com.ibm.rdm.web.notifications.EventTopics.OBJECT_TYPE_CHANGED],
			callbackFunction: this._onArtifactTypeChangedNotify,
			callbackScope: this,
			neverDelayCallback: true
		});

	},

	///////////////////////////////////////////////////////////////////////////
	// Public API

	// Functions to support ProcessOperationsCache._isSaveTeamSecuritySpecificationAllowed ...

	setSaveTeamSecuritySpecificationDeferred: function(processAreaURI, deferred) {
		this.mapSaveTeamSecurityToDeferred[processAreaURI] = deferred;
	},

	getSaveTeamSecuritySpecificationDeferred: function(processAreaURI) {
		return this.mapSaveTeamSecurityToDeferred[processAreaURI];
	},
	
	setQuickPrintDeferred: function(processAreaURI, deferred) {
		this.mapQuickPrintToDeferred[processAreaURI] = deferred;
	},

	getQuickPrintDeferred: function(processAreaURI) {
		return this.mapQuickPrintToDeferred[processAreaURI];
	},	


	setAcquireManualLockPermissionDeferred: function(processAreaURI, deferred) {
		this.mapAcquireManualLockDeferred[processAreaURI] = deferred;
	},
	
	getAcquireManualLockPermissionDeferred: function(processAreaURI) {
	   return this.mapAcquireManualLockDeferred[processAreaURI];  
	},

	// Functions to support ProcessOperationsCache._getFolderPermissions ...

	setFolderPermissionsDeferred: function(processAreaURI, deferred) {
		this.mapFolderProcessAreaToDeferred[processAreaURI] = deferred;
	},

	getFolderPermissionsDeferred: function(processAreaURI) {
		return this.mapFolderProcessAreaToDeferred[processAreaURI];
	},
	
	// Functions to support ProcessOperationsCache._getModifyLocksPermissions ...

	setModifyLocksPermissionsDeferred: function(processAreaURI, deferred) {
		this.mapModifyLocksProcessAreaToDeferred[processAreaURI] = deferred;
	},

	getModifyLocksPermissionsDeferred: function(processAreaURI) {
		return this.mapModifyLocksProcessAreaToDeferred[processAreaURI];
	},		


	// Functions to support ProcessOperationsCache._isDeleteInstanceOfArtifactTypeAllowed ...

	setDeleteInstanceOfArtifactTypeAllowedDeferred: function(
		processAreaURI, artifactTypeActionId, deferred)
	{
		var key = this._getPairKey(processAreaURI, artifactTypeActionId);
		this.mapDeleteInstanceToDeferred[key] = deferred;
	},

	getDeleteInstanceOfArtifactTypeAllowedDeferred: function(
		processAreaURI, artifactTypeActionId)
	{
		var key = this._getPairKey(processAreaURI, artifactTypeActionId);
		return this.mapDeleteInstanceToDeferred[key];
	},


	// Functions to support ProcessOperationsCache._isCreateInstanceOfArtifactTypeAllowed ...

	setCreateInstanceOfArtifactTypeAllowedDeferred: function(
		processAreaURI, artifactTypeActionId, deferred)
	{
		var key = this._getPairKey(processAreaURI, artifactTypeActionId);
		this.mapCreateInstanceToDeferred[key] = deferred;
	},

	getCreateInstanceOfArtifactTypeAllowedDeferred: function(
		processAreaURI, artifactTypeActionId)
	{
		var key = this._getPairKey(processAreaURI, artifactTypeActionId);
		return this.mapCreateInstanceToDeferred[key];
	},


	// Functions to support ProcessOperationsCache._getModifyArtifactPermissions ...

	setModifyArtifactPermissionsDeferred: function(
		processAreaURI, artifactTypeActionId, deferred)
	{
		var key = this._getPairKey(processAreaURI, artifactTypeActionId);
		this.mapModifyArtifactTypeToDeferred[key] = deferred;
	},

	getModifyArtifactPermissionsDeferred: function(
		processAreaURI, artifactTypeActionId)
	{
		var key = this._getPairKey(processAreaURI, artifactTypeActionId);
		return this.mapModifyArtifactTypeToDeferred[key];
	},


	// Functions to support ProcessOperationsCache._isModifyLinkAllowed ...

	setModifyLinkDeferred: function(
		processAreaURI, linkTypeActionId, linkTypePermission, deferred)
	{
		var key = this._getTripleKey(processAreaURI, linkTypeActionId, linkTypePermission);
		this.mapModifyLinkInstanceToDeferred[key] = deferred;
	},

	getModifyLinkDeferred: function(processAreaURI, linkTypeActionId, linkTypePermission) {
		var key = this._getTripleKey(processAreaURI, linkTypeActionId, linkTypePermission);
		return this.mapModifyLinkInstanceToDeferred[key];
	},


	// Functions to support ProcessOperationsCache._getPermissibleLinkTypesForCreation ...

	setPermissibleLinkTypesForCreationDeferred: function(processAreaURI, deferred) {
		this.mapPermissibleLinkTypesForCreationToDeferred[processAreaURI] =
			deferred;
	},

	getPermissibleLinkTypesForCreationDeferred: function(processAreaURI) {
		return this.mapPermissibleLinkTypesForCreationToDeferred[processAreaURI];
	},
	

	// Functions to support ProcessOperationsCache._isCreateLinkAllowed ...

	setCreateLinkDeferred: function(
		processAreaURI, linkTypeActionId, deferred)
	{
		var key = this._getPairKey(processAreaURI, linkTypeActionId);
		this.mapCreateLinkInstanceToDeferred[key] =
			deferred;
	},

	getCreateLinkDeferred: function(processAreaURI, linkTypeActionId) {
		var key = this._getPairKey(processAreaURI, linkTypeActionId);
		return this.mapCreateLinkInstanceToDeferred[key];
	},
	

	// Functions to support ProcessOperationsCache._isDeleteLinkAllowed ...

	setDeleteLinkDeferred: function(
		processAreaURI, linkTypeActionId, deferred)
	{
		var key = this._getPairKey(processAreaURI, linkTypeActionId);
		this.mapDeleteLinkInstanceToDeferred[key] =
			deferred;
	},

	getDeleteLinkDeferred: function(processAreaURI, linkTypeActionId) {
		var key = this._getPairKey(processAreaURI, linkTypeActionId);
		return this.mapDeleteLinkInstanceToDeferred[key];
	},
	

	///////////////////////////////////////////////////////////////////////////
	// Private API

	_getPairKey: function(processAreaURI, artifactTypeActionId) {
		return (processAreaURI + artifactTypeActionId);
	},

	_getTripleKey: function(processAreaURI, artifactTypeActionId, linkTypePermission) {
		return (processAreaURI + artifactTypeActionId + linkTypePermission);
	},
	
	_onLinkTypeAddedNotify: function(eventParams) {
		this.mapPermissibleLinkTypesForCreationToDeferred = {};
	},
	
	//#68914 Attributes added to artifact are shown but can't be edited
	_onArtifactTypeChangedNotify: function(eventParams) {
		if (eventParams.notifierParams.projectURI) {
			com.ibm.rdm.web.server.cache.ProcessOperationsCache.artifactTypeChangedNotify(eventParams, this);
		}	
	}
});


})();
